import rx
from fs.osfs import OSFS
from fs import path as fp
from fs import copy as fc


def files(fs, filters=None, walk=False, exclude_files=None):
    if filters is None:
        filters = []
    if exclude_files is None:
        exclude_files = []
    if walk:
        return (rx.Observable
                .from_(fs.walk.files('.', filter=filters))
                .map(lambda p: fp.relativefrom(fs.getsyspath('.'), p)))
    else:
        return rx.Observable.from_(fs.filterdir('.', files=filters, exclude_files=exclude_files, exclude_dirs=['*'])).map(lambda info: info.name)


def dirs(fs, filters=None, walk=False):
    if filters is None:
        filters = []
    if walk:
        return (rx.Observable
                .from_(fs.walk.dirs('.', filter=filters))
                .map(lambda p: fp.relativefrom(fs.getsyspath('.'), p)))
    else:
        return rx.Observable.from_(fs.filterdir('.', exclude_files=['*'], dirs=filters)).map(lambda info: info.name)


def make_subdirs(fs, prefix, nb_dirs):
    for i in range(nb_dirs):
        fs.mkdir('{0}.{1}'.format(prefix, i))


class Mapper:
    def __init__(self, fs, filters=None):
        self.fs = fs
        self.root = fp.normpath(self.fs.getsyspath('.'))
        if filters is None:
            filters = ['sub*']
        self.filters = filters

        self.subdirs = []
        dirs(self.fs, self.filters).subscribe(self.subdirs.append)

    def id_dir(self, path):
        if not fp.normpath(path) in self.subdirs:
            return None
        suffix = fp.splitext(path)
        return int(suffix[1:])

    def dirs(self):
        return rx.Observable.from_(self.subdirs)

    def subdirs_with_id(self):
        dir_with_id = [(d, self.id_dir(d)) for d in self.subdirs]
        return rx.Observable.from_(dir_with_id)

    def files(self, filename):
        return (self.dirs()
                .map(lambda d: fp.combine(d, filename)))



#!/home/hongxwing/anaconda3/bin/python
import click
import yaml
from fs.osfs import OSFS
import os
import subprocess
import sys
from pygate import service

DEFAULT_CONFIG_FILE = service.ConfigMaker.DEFAULT_CONFIG_FILENAME


@click.group()
def gate():
    pass


@gate.command()
@click.option('--target', '-t', type=str, default='.', help='target directory')
@click.option('--config', '-c', type=str, default=DEFAULT_CONFIG_FILE, help='config file name')
def make_config(target, config):
    service.ConfigMaker.make(target, config)


def load_config(config):
    with open(config) as fin:
        return yaml.load(fin)


@gate.command()
@click.option('--config', '-c', type=str, default=DEFAULT_CONFIG_FILE, help='config YAML file name')
@click.option('--templates', '-t', 'content', flag_value='templates', help='copy files from template directory')
@click.option('--shell', '-s', 'content', flag_value='shell', help='generate shell files')
@click.option('--dirs', '-d', 'content', flag_value='dirs', help='generate and copy files to sub directories')
@click.option('--macss', '-m', 'content', flag_value='macs', help='generate mac files')
@click.option('--all', '-a', 'content', flag_value='all',
              default=True, help='All tasks above')
def init(config, content):
    """
    Initialize working directory
    """
    c = load_config(config)
    make_all = False
    if content == 'all':
        make_all = True
    if content == 'macs' or make_all:
        service.Initializer.mac(c)
    if content == 'templates' or make_all:
        service.Initializer.templates(c)
    if content == 'shell' or make_all:
        service.ShellMaker.run(c)
        service.ShellMaker.post(c)
        # service.make_run_sh(c['target'], c['run_sh'],
        #                     c['main_mac'], c['analysis_c'])
        # service.make_post_sh(c['target'], c['post_sh'])
    if content == 'dirs' or make_all:
        service.SubDirectoryMaker.make_subs(c)


@click.command()
@click.option('--config', '-c', type=str, default=DEFAULT_CONFIG_FILE, help='config file name')
def run(config):
    c = load_config(config)
    service.run(c['target'], c['main_mac'], stdout=c['stdout'])


@gate.command()
@click.option('--config', '-c', type=str, default=DEFAULT_CONFIG_FILE, help='config YAML file name')
def submit(config):
    # TODO: add submit service
    c = load_config(config)
    service.Submitter.submit(c)


@gate.command()
@click.option('--config', '-c', type=str, default=DEFAULT_CONFIG_FILE, help='config YAML file name')
def merge(config):
    c = load_config(config)
    service.Merger.merge(c)


@gate.command()
@click.option('--config', '-c', type=str, default=DEFAULT_CONFIG_FILE, help='config YAML file name')
@click.option('--dirs', '-d', 'content', flag_value='dirs', default=True)
@click.option('--all', '-a', 'content', flag_value='all')
@click.option('--dryrun', is_flag=True)
def clear(config, content, dryrun):
    c = load_config(config)
    if content == 'dirs':
        service.Cleaner.subs(c, dryrun)
    elif content == 'all':
        service.Cleaner.all(c, dryrun)


if __name__ == "__main__":
    gate()



import unittest
# from fs.memoryfs import MemoryFS as FS
from fs.tempfs import TempFS as FS
from pygate import methods


class TestMethods(unittest.TestCase):
    def setUp(self):
        self.fs = FS()
        for i in range(5):
            self.fs.makedir('sub.{}'.format(i))
        for i in range(3):
            self.fs.touch('test{}.mac'.format(i))
        for i in range(3):
            self.fs.touch('result{}.txt'.format(i))
        self.fs.makedir('testdir')
        self.fs.touch('testdir/tmp.txt')

    def tearDown(self):
        self.fs.close()

    def test_files(self):
        expected = set()
        for i in range(3):
            expected.add('test{}.mac'.format(i))
            expected.add('result{}.txt'.format(i))
        result = set()
        methods.files(self.fs).subscribe(result.add)
        self.assertEqual(result, expected)

    def test_files_exclude(self):
        expected = {'test{}.mac'.format(i) for i in range(3)}
        result = set()
        methods.files(self.fs, exclude_files=['result*']).subscribe(result.add)
        self.assertEqual(result, expected)

    def test_subdirs(self):
        expected = {'sub.{}'.format(i) for i in range(5)}
        result = set()
        methods.dirs(self.fs, filters=['sub*']).subscribe(result.add)
        self.assertEqual(result, expected)

    def test_mapper_subdirs(self):
        m = methods.Mapper(self.fs)
        result = set()
        m.files('result.txt').subscribe(result.add)
        expected = {'sub.{}/result.txt'.format(i) for i in range(5)}
        self.assertEqual(result, expected)



from fs.osfs import OSFS
# from . import methods


from .template.service import MacMaker


# class EnsureFS:
#     def __init__(self, fs):
#         if isinstance(fs, str):
#             self.fs = OSFS(fs)
#             self.need_close = True
#         else:
#             self.fs = fs
#             self.need_close = False

#     def __enter__(self):
#         return self.fs

#     def __exit__(self, type, value, trackback):
#         if self.need_close:
#             self.fs.close()


# def copy_file(source, target, name):
#     """
#     Inputs:
#         source: directory, str or fs.baseFS
#         target: directory, str or fs.baseFS
#         name: filename, str
#     Raises:
#         None
#     Returns:
#         None
#     """
#     with EnsureFS(source) as fss, EnsureFS(target) as fst:
#         fc.copy_file(fss, name, fst, name)


# def copy_dir(source, target, filters=None):
#     with EnsureFS(source) as fss, EnsureFS(target) as fst:
#         filters = filters or ['*']
#         (methods.files(fss, filters)
#             .subscribe(lambda f: copy_file(fss, fst, f)))


# def copy_group(source, target, group_name):
#     with EnsureFS(source) as fss, EnsureFS(target) as fst:
#         with fss.opendir(group_name) as sd:
#             copy_dir(sd, fst)


def make_config(target='.'):
    from .config_maker import ConfigMaker
    from .config import config as c
    with OSFS('.') as fs:
        ConfigMaker.make(fs, '.')


def init(config, method):
    from .initializer import Initializer
    with OSFS('.') as fs:
        initer = Initializer(fs, config)
        if method == 'pre' or method == 'all':
            initer.pre_sub()
        if method == 'sub' or method == 'all':
            initer.make_sub()


def submit(config):
    from .submitter import Submitter
    with OSFS('.') as fs:
        Submitter(fs, config).submit()


def merge(config):
    from .merger import Merger
    with OSFS('.') as fs:
        Merger(fs, config).merge()


def clean(config):
    from .cleaner import Cleaner
    with OSFS('.') as fs:
        Cleaner(fs, config).clean()

# class Merger:
#     @classmethod
#     def hadd(cls, fs, target, filenames, dryrun=False):
#         path_target = fs.getsyspath(target)
#         path_filenames = [fs.getsyspath(f) for f in filenames if fs.exists(f)]
#         call_args = ['hadd', path_target] + path_filenames
#         if dryrun:
#             print(' '.join(call_args))
#         else:
#             with subprocess.Popen(call_args,
#                                   stdout=subprocess.PIPE,
#                                   stderr=subprocess.PIPE) as p:
#                 sys.stdout.write(p.stdout.read().decode())
#                 sys.stderr.write(p.stderr.read().decode())

#     @classmethod
#     def direct(cls, fs, target, filenames):
#         with fs.open(targe, 'w') as fout:
#             for f in filenames:
#                 with fs.open(f) as fin:
#                     print(fin.read(), end='', file=fout)

#     @classmethod
#     def merge(cls, c):
#         method = (c.get('merge') or {}).get('method') or 'DIRECT'
#         files_to_merge = ((c.get('merge') or {}).get('filenames') or [])
#         with OSFS(c['target']) as fs:
#             for mfn in files_to_merge:
#                 filenames = []
#                 (methods.Mapper(fs).files(mfn)
#                     .subscribe(filenames.append))
#                 if method.upper() == 'HADD':
#                     cls.hadd(fs, mfn, filenames, c.get('dryrun'))
#                 elif method.upper() == 'DIRECT':
#                     cls.direct(fs, mfn, filenames)



import unittest
from ruamel.yaml import YAML
from fs.memoryfs import MemoryFS
from fs.osfs import OSFS
from dxpy.filesystem import Path
import jinja2
from pygate import shell


class TestShellScript(unittest.TestCase):
    def test_shell_bash(self):
        from fs.osfs import OSFS
        result = shell.make(OSFS('/tmp'), '/sub.1', 'map', [{'method':'Gate', 'filename': 'main.mac'}], shell='bash')
        from pygate.utils import load_script
        expect = load_script('map_bash_sample.sh')
        self.assertEqual(result, expect)
        




from dxpy import batch
class Merger:
    def __init__(self, config, root_fs):
        """
        config: a dict of configurations:
            tasks: list of tasks:
                method: 'HADD' | 'CAT' | 'SUM'
                files: list of filenames
            dryrun: [OPTIONAL] bool
            target: [OPTIONAL] str, relative path of output directory
            verbose: [OPTIONAL] int
        """
        self.c = config
        self.fs = root_fs
        if 'dryrun' in self.c:
            self.dryrun = self.c['dryrun']
        else:
            self.dryrun = False
        self.target = self.get(['target']) or '.'
        self.verbose = self.get(['verbose']) or 1

    def func(self, name):
        return {
            'HADD': self.hadd,
            'CAT': self.cat,
            'SUM': self.sum
        }[name.upper()]

    def run(self):
        if not 'task' in self.c:
            return
        for task in self.c['tasks']:
            for fn in task['files']:
                self.func(task['method'])(fn)

    @classmethod
    def hadd(cls, filenames):
        path_target = fs.getsyspath(target)
        path_filenames = [fs.getsyspath(f) for f in filenames if fs.exists(f)]
        call_args = ['hadd', path_target] + path_filenames
        if dryrun:
            print(' '.join(call_args))
        else:
            with subprocess.Popen(call_args,
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.PIPE) as p:
                sys.stdout.write(p.stdout.read().decode())
                sys.stderr.write(p.stderr.read().decode())

    @classmethod
    def direct(cls, fs, target, filenames):
        with fs.open(targe, 'w') as fout:
            for f in filenames:
                with fs.open(f) as fin:
                    print(fin.read(), end='', file=fout)

    @classmethod
    def merge(cls, c):
        method = (c.get('merge') or {}).get('method') or 'DIRECT'
        files_to_merge = ((c.get('merge') or {}).get('filenames') or [])
        with OSFS(c['target']) as fs:
            for mfn in files_to_merge:
                filenames = []
                (methods.Mapper(fs).files(mfn)
                    .subscribe(filenames.append))
                if method.upper() == 'HADD':
                    cls.hadd(fs, mfn, filenames, c.get('dryrun'))
                elif method.upper() == 'DIRECT':
                    cls.direct(fs, mfn, filenames)



# import sys
# import warnings
# from fs.osfs import OSFS
# from collections import OrderedDict


# class Defaults:
#     SHELL = 'zsh'


# class Snippets:
#     SHELL = {
#         'bash': r'#!/bin/bash',
#         'zsh': r"#!/usr/bin/zsh"
#     }

#     SOURCE = {
#         'bash': r'~/.bashrc',
#         'zsh': r'~/.zshrc'
#     }

#     SBATCH = ["#SBATCH -o %J.out",
#               "#SBATCH -e %J.err"]

#     @classmethod
#     def add_task_func(cls, key):
#         TASK_MAPPING = {
#             'GATE': cls.add_gate,
#             'ROOT': cls.add_root,
#             'HADD': cls.add_hadd,
#             'CLEARSUB': cls.add_clear_sub,
#             'MERGE': cls.add_merge}
#         return TASK_MAPPING.get(key.upper())

#     @classmethod
#     def add_shell(cls, scripts, name):
#         name = name.lower()
#         if not name in cls.SHELL:
#             fmt = 'Unknown shell name {0}, use {1} instead.'
#             warnings.warn(fmt.format(name, Defaults.SHELL), RuntimeWarning)
#             name = Defaults.SHELL
#         scripts.append(cls.SHELL[name])

#     @classmethod
#     def add_source(cls, scripts, name):
#         name = name.lower()
#         if not name in cls.SHELL:
#             fmt = 'Unknown shell name {0}, use {1} instead.'
#             warnings.warn(fmt.format(name, Defaults.SHELL), RuntimeWarning)
#             name = Defaults.SHELL
#         scripts.append('source ' + cls.SOURCE[name])

#     @classmethod
#     def add_sbatch(cls, scripts):
#         for s in cls.SBATCH:
#             scripts.append(s)

#     @classmethod
#     def add_gate(cls, scripts, filename):
#         scripts.append('Gate {0}'.format(filename))

#     @classmethod
#     def add_root(cls, scripts, filename):
#         scripts.append('root -q -b {0}'.format(filename))

#     @classmethod
#     def add_hadd(cls, scripts, target, filenames):
#         if isinstance(filenames, str):
#             filenames = [filenames]
#         scripts.append('hadd {0} '.format(target) + ' '.join(filenames))

#     @classmethod
#     def add_merge(cls, scripts):
#         scripts.append('pygate merge')

#     @classmethod
#     def add_clear_sub(cls, scripts):
#         scripts.append('pygate clear --dirs')

#     @classmethod
#     def add_task(cls, scripts, task_name, payloads):
#         add_func = cls.add_task_func(task_name)
#         if add_func is None:
#             fmt = 'Unknown task name {0}, skipped.'
#             warnings.warn(fmt.format(task_name), RuntimeWarning)
#             return
#         payloads = payloads or {}
#         add_func(scripts, **payloads)


# def _add_hadd(commands, fs, workdir, task):
#     from dxpy import batch
#     if task['method'] == 'hadd':
#         with fs.opendir(workdir) as d:
#             files = batch.files_in_directories(
#                 d, task['sub_dirs'], [task['filename']])
#             paths = [d.getsyspath(p) for p in files]
#             paths = [fs.getsyspath(p) for p in paths]
#         commands.append('hadd {0}'.format(
#             task['filename']) + ' '.join(paths))


# def _add_sum():
#     pass


# MAP_TASKS = [_add_gate, _add_root]


# def _make_commands(fs, workdir, task_type, tasks):
#     commands = ['']
#     if task_type == 'map':
#         for t in tasks:
#             for add_cmd in MAP_TASKS:
#                 add_cmd(commands, fs, workdir, t)
#     elif task_type == 'merge':
#         pass
#     commands.append('')
#     return "\n".join(commands)


# def _make_script_file(fs, workdir: str, output: str, task_type: str, tasks: list, shell='zsh'):
#     scrpt_tpl = __load_shell_template(task_type, shell)
#     p_ser, p_loc = _get_workdir_on_local_and_server(fs.getsyspath(workdir))
#     s = scrpt_tpl.render(local_work_directory=p_loc,
#                          server_work_directory=p_ser,
#                          commands=_make_commands(fs, workdir, task_type, tasks))
#     with fs.opendir(workdir) as d:
#         with d.open(output, 'w') as fout:
#             print(s, file=fout)


class ShellScriptBase:
    def __init__(self, fs, workdir: str, output: str, tasks: list, shell='zsh'):
        self.fs = fs
        self.workdir = workdir
        self.output = output
        self.tasks = tasks
        self.shell = shell.lower()
        if not self.shell in ['zsh', 'bash']:
            raise ValueError("Unknown shell {}.".format(self.shell))
        self.content = None
        self.task_type = None

    def _load_shell_template(self):
        from .utils import load_script
        import jinja2
        script_name = "{0}_{1}.sh".format(self.task_type, self.shell)
        return jinja2.Template(load_script(script_name))

    def _make(self):
        raise NotImplemented

    def dump(self):
        self._make()
        with self.fs.opendir(self.workdir) as d:
            with d.open(self.output, 'w') as fout:
                print(self.content, file=fout)


class ShellScriptMap(ShellScriptBase):
    def __init__(self, fs, workdir: str, output: str, tasks: list, shell='zsh'):
        super(__class__, self).__init__(fs, workdir, output, tasks, shell)
        self.task_type = 'map'
        self.commands = []

    def _add_if_gate(self, task):
        if task.get('method') == 'Gate':
            self.commands.append("Gate {}".format(task['filename']))

    def _add_if_root(self, task):
        if task['method'] == 'root':
            self.commands.append("root -q -b {}".format(task['filename']))

    def _get_workdir_on_local_and_server(self):
        from dxpy.filesystem import Path
        p_ser = self.fs.getsyspath(Path(self.workdir).abs)
        p_loc = (Path('~/Slurm/') / Path(self.workdir).rel).rel
        return p_ser, p_loc

    def _make(self):
        scrpt_tpl = self._load_shell_template()
        supported_tasks = [self._add_if_gate, self._add_if_root]
        for t in self.tasks:
            for add_if_func in supported_tasks:
                add_if_func(t)
        p_ser, p_loc = self._get_workdir_on_local_and_server()
        self.content = scrpt_tpl.render(local_work_directory=p_loc,
                                        server_work_directory=p_ser,
                                        commands='\n'.join(self.commands))


class ShellScriptMerge(ShellScriptBase):
    def __init__(self, fs, workdir: str, output: str, tasks: list, shell='zsh'):
        super(__class__, self).__init__(fs, workdir, output, tasks, shell)
        self.task_type = 'merge'

    def _make(self):
        scrpt_tpl = self._load_shell_template()
        self.content = scrpt_tpl.render(merge='merge' in self.tasks,
                                        clean='clean' in self.tasks)


class ShellScriptMaker:
    def __init__(self, fs, config):
        from .utils import sub_dir_filters
        self.fs = fs
        self.c = config['shell']
        self.sub_filters = sub_dir_filters(config)

    def make(self):
        from dxpy import batch
        sub_dirs = batch.DirectoriesFilter(self.sub_filters).lst(self.fs)
        for d in sub_dirs:
            ShellScriptMap(self.fs, d, self.c['map']['output'],
                           self.c['map']['tasks'], self.c['type']).dump()
        ShellScriptMerge(self.fs, '.', self.c['merge']['output'],
                         self.c['merge']['tasks'], self.c['type']).dump()


# class ShellScript:
#     def __init__(self, workdir=None, task_type='map', tasks=None, shell='zsh'):
#         self.workdir = workdir
#         if not task_type.lower() in ['map', 'merge']:
#             raise ValueError(
#                 "Task type {} is not supported.".format(task_type))
#         self.task_type = task_type.lower()
#         if not shell.lower() in ['bash', 'zsh']:
#             raise ValueError("Shell {} is not supported.".format(task_type))
#         self.shell = shell.lower()
#         self.tasks = tasks or []
#         self._make()

#     def dump(self):
#         if isinstance(self.output, str):
#             with OSFS('.') as fs:
#                 with fs.open(self.output, 'w') as fout:
#                     print('\n'.join(self.scripts), file=fout)
#         else:
#             print('\n'.join(self.scripts), file=self.output)

#     def add_head(self):
#         Snippets.add_shell(self.scripts, self.shell)
#         Snippets.add_sbatch(self.scripts)
#         Snippets.add_source(self.scripts, self.shell)
#         self.scripts.append('hostname')
#         self.scripts.append('date')

#     def add_tasks(self):
#         for t in self.tasks:
#             Snippets.add_task(self.scripts, t.get('name'), t.get('payloads'))

#     def add_tail(self):
#         self.scripts.append('date')
#         self.scripts.append('')
